/*  ir3.c
 *
 *  A simple utility for Linux to send and receive LNP message
 *  between PC and RCX. Quite inefficient, works only in slow
 *  mode for me. Tested with BrickOS 0.2.6.10.6 and Debian Linux.
 *
 *  The code is blend of original LegOS-BrickOS firmdl3.c (ver 3.1) and
 *  ir.c - the same utility for Windows.
 *
 *  Not sure how this still works with USB tower.
 *  Addressing LNP not supported.
 *  All hints and comments will be appreciated.
 *
 *  The firmdl3.c licence follows:
 *
 *  -----------------------------------------begin-firmdl3.c-licence-----
 *  firmdl3.c
 *
 *  A firmware downloader for the RCX.  Version 3.0.  Supports single and
 *  quad speed downloading.
 *
 *  The contents of this file are subject to the Mozilla Public License
 *  Version 1.0 (the "License"); you may not use this file except in
 *  compliance with the License. You may obtain a copy of the License at
 *  http://www.mozilla.org/MPL/
 *
 *  Software distributed under the License is distributed on an "AS IS"
 *  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 *  License for the specific language governing rights and limitations
 *  under the License.
 *
 *  The Original Code is Firmdl code, released October 3, 1998.
 *
 *  The Initial Developer of the Original Code is Kekoa Proudfoot.
 *  Portions created by Kekoa Proudfoot are Copyright (C) 1998, 1999
 *  Kekoa Proudfoot. All Rights Reserved.
 *
 *  Contributor(s): Kekoa Proudfoot <kekoa@graphics.stanford.edu>
 *                  Laurent Demailly
 *                  Allen Martin
 *                  Markus Noga
 *                  Gavin Smyth
 *                  Luis Villa
 *
 *  ------------------------------------------end-firmdl3.c-licence-----
 *  modifications for ir3: Pavel Petrovic, ppetrovic@acm.org
 */
#if defined(_WIN32)
  #include <windows.h>
#endif


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include <time.h>

#include "rcx_comm.h"

#define DEFAULTTTY   "/dev/ttyS0" /* Linux - COM1 */
#define DEFAULTFILENAME     "/dev/null"

// state of the LNP protocol when receiving a packet
typedef enum { LNPwaitHeader, LNPwaitLength, LNPwaitData, LNPwaitCRC } lnp_integrity_state_t;

/* Global variables */

int tty_usb = 0;
extern int __comm_debug;
static char packet[257];
static char *progname;
static int receive_timeout;
static int print;
static char *filename = DEFAULTFILENAME;

/* Defines */

#define BUFFERSIZE      4000
#define WAKEUP_TIMEOUT  4000


// modified originally from LegOS sources
unsigned char lnp_checksum( const unsigned char *data, unsigned length )
{
  unsigned char a = 0xff;
  while (length > 0) {
    a = a + *data;
    data++;
    length--;
  }
  return a;
}

//send msg as LNP packet
int msg_send(FILEDESCR fd, unsigned char *msg, int len)
{
    unsigned char send[BUFFERSIZE];
    struct timespec t;

    send[0] = 0xff;
    mywrite(fd, send, 1);

    t.tv_sec = 0;
    t.tv_nsec = 50000000;
    nanosleep(&t, 0);

    send[0] = 0xf0;
    send[1] = len + 1;
    memcpy(send + 2, msg, len);
    send[len + 2] = (char)0;
    send[len + 3] = (char)lnp_checksum(send, len + 3);

    if (len + 4 != mywrite(fd, send, len + 4))
    {
      perror("error writing to port");
      return 0;
    }

    return 1;
}

//send file as LNP packet
int file_send(FILEDESCR fd, unsigned char *filename)
{
    int f;
    int filelen, rv;
    char *msg;

    f = open(filename, O_RDONLY);
    if (f == -1) {
      fprintf(stderr, "filename:%s\n", filename);
      perror("can't open message file");
      return 0;
    }

    if ((off_t)-1 == (filelen = lseek(f, 0, SEEK_END)))
    {
      perror("can't access message file");
      return 0;
    }

    lseek(f, 0, SEEK_SET);
    msg = (char *)malloc(filelen + 2);
    if (!msg)
    {
      perror("not enough mem?");
      return 0;
    }
    if (filelen != read(f, msg, filelen))
    {
      perror("error reading msg file");
      return 0;
    }
    close(f);

    rv = msg_send(fd, msg, filelen);
    free(msg);
    return rv;
}

// LNP, bit modified from LegOS sources...
int lnp_integrity_byte(unsigned char b)
{
  static unsigned char buffer[259];
  static int bytesRead,endOfData;
  static int lnp_integrity_state = LNPwaitHeader;

  if(lnp_integrity_state==LNPwaitHeader)
  bytesRead=0;

  buffer[bytesRead++]=b;

//  printf("{%d}", b);

  switch(lnp_integrity_state)
  {
    case LNPwaitHeader:
        // valid headers are 0xf0 .. 0xf7
        if((b & (unsigned char) 0xf8) == (unsigned char) 0xf0)
          lnp_integrity_state++;
        break;

    case LNPwaitLength:
        endOfData=b+2;
        lnp_integrity_state++;
        break;

    case LNPwaitData:
        if(bytesRead==endOfData)
          lnp_integrity_state++;
        break;

    case LNPwaitCRC:
        if(b == (unsigned char)lnp_checksum(buffer,endOfData))
        {
           buffer[buffer[1]+2]=0;
           strcpy(packet,buffer+2);
           lnp_integrity_state=LNPwaitHeader;
           return 1;
        }
        else
	{
	  printf("bad checksum\n");
	  lnp_integrity_state=LNPwaitHeader;
	}
  }
  return 0;
}

//receive next LNP message to global array packet[]
int msg_receive(FILEDESCR fd)
{
  int status, len, i, j, end_of_message;
  unsigned char recv[BUFFERSIZE + 1];
  FILE *g;

  packet[0] = 0;

  // USB tower does not need to be waken up...
  if(tty_usb==0) {
    if ((status = rcx_wakeup_tower(fd, WAKEUP_TIMEOUT)) < 0) {
      fprintf(stderr, "%s: %s\n", progname, rcx_strerror(status));
      return 0;
    }
  }

  j = 0;
  do {
    len = nbread(fd, recv, BUFFERSIZE, receive_timeout);
    // process all received bytes through LNP

    i = 0; j++;
    do {
      end_of_message = 0;
      for (;i < len && !end_of_message; i++)
        if (lnp_integrity_byte(recv[i])) end_of_message = 1;
      if (print && end_of_message)
      {
	    printf("%s\n", packet);
	    g = fopen(filename, "a+");
        if (g) { fprintf(g, "%s\n", packet); fclose(g); }
        if (strcmp(packet, "\bye") == 0) print = 0;
	    packet[0] = 0;
      }
    } while (print && (i < len));

    if (tty_usb == 0)
    {
      if (print && (j * receive_timeout > 3000))
      {
        j = 0;
        recv[0] = 0x10;
        mywrite(fd, recv, 1);
      }
    }
#ifdef _WIN32
	FlushFileBuffers(fd);
#endif

  } while (print);

  return end_of_message;
}


//receive next LNP message and save it to file
int file_receive(FILEDESCR fd, char *filename)
{
  int f;

  if (!msg_receive(fd))
  {
    printf("no msg received\n");
    return 0;
  }

  // write message to file
  if (-1 == (f = open(filename, O_WRONLY | O_CREAT | O_TRUNC , S_IREAD | S_IWRITE)))
  {
    perror("openning output file");
    return 0;
  }
  if (strlen(packet) != write(f, packet, strlen(packet)))
  {
    perror("writing packet to file");
    return 0;
  }
  if (-1 == close(f))
  {
    perror("closing output file");
    return 0;
  }

  return 1;
}


int main (int argc, char **argv)
{
    char *tty = NULL;
    int use_fast = 0;
    int usage = 0;
    int receive = 0;
    FILEDESCR fd;

    /* Parse command line - from firmdl.c */


    print = 0;
    receive_timeout = 100;
    progname = argv[0];
    argv++; argc--;
    while (argc && argv[0][0] == '-') {
	if (argv[0][1] == '-') {
	    if (!strcmp(argv[0], "--")) {
		argv++; argc--;
		break;
	    }
	    else if (!strcmp(argv[0], "--debug")) {
		__comm_debug = 1;
	    }
	    else if (!strcmp(argv[0], "--fast")) {
		use_fast = 1;
	    }
	    else if (!strcmp(argv[0], "--slow")) {
		use_fast = 0;
	    }
	    else if (!strcmp(argv[0], "--print")) {
		print = 1;
		receive = 1;
	    }
	    else if (!strcmp(argv[0], "--receive")) {
		receive = 1;
		if (argc < 2) usage = 1;
		else {
		  sscanf(argv[1], "%d", &receive_timeout);
		  argc--;
		  argv++;
		}
	    }
	    else if (!strncmp(argv[0], "--tty", 5)) {
		if (argv[0][5] == '=') {
		    tty = &argv[0][6];
		}
		else if (argc > 1) {
		    argv++; argc--;
		    tty = argv[0];
		}
		else
		    tty = "";
		if (!tty[0]) {
		    fprintf(stderr, "%s: invalid tty: %s\n", progname, tty);
		    exit(1);
		}
	    }
	    else if (!strcmp(argv[0], "--help")) {
		usage = 1;
	    }
	    else {
		fprintf(stderr, "%s: unrecognized option %s\n",
			progname, argv[0]);
		exit(1);
	    }
	}
	else {
	    char *p = &argv[0][1];
	    if (!*p)
		break;
	    while (*p) {
		switch (*p) {
		case 'f': use_fast = 1; break;
		case 's': use_fast = 0; break;
		case 'h': usage = 1; break;
		case 'p': print = 1; receive = 1; break;
	        case 'r': receive = 1;
			  if (argc < 2) usage = 1;
			  else {
		            sscanf(argv[1], "%d", &receive_timeout);
			    argc--;
			    argv++;
			  }
			  break;

		default:
		    fprintf(stderr, "%s: unrecognized option -- %c\n",
			    progname, *p);
		    exit(1);
		}
		p++;
	    }
	}
	argv++;
	argc--;
    }

    if (usage || ((argc != 1) && !(print && receive))) {
	char *usage_string =
	    "      --debug      show debug output, mostly raw bytes\n"
	    "  -f, --fast       use fast 4x downloading\n"
	    "  -s, --slow       use slow 1x downloading (default)\n"
	    "  -r timeout,\n"
	    "      --receive    receive a single message and quit\n"
	    "                   (if -r or -p not specified, sending)\n"
	    "  -p  --print      print mode: all received messages printed\n"
	    "                   to stdout (or file, if specified) until\n"
	    "                   '\bye' message or signal kills the program\n"
	    "                   if both -r and -p specified, receive all msgs.\n"
	    "      --tty=TTY    assume tower connected to TTY\n"
#if defined(_WIN32)
	    "      --tty=usb    assume tower connected to USB\n"
#endif
	    "  -h, --help       display this help and exit\n"
	    ;

	fprintf(stderr, "usage: %s [options] filename\n", progname);
	fprintf(stderr, usage_string);
	exit(1);
    }

    if (!tty) tty = getenv("RCXTTY");  /* Get the tty name */
    if (!tty) tty = DEFAULTTTY;

#if defined(_WIN32)
	    //Check the command line to see if IR tower is USB.
	    if(stricmp(tty,"usb")==0)	{
		tty_usb = 1;
		if(__comm_debug)
			fprintf(stderr, "Hary Mahesan - USB IR Tower Mode.\n");
		tty = "\\\\.\\legotower1"; // Set the correct usb tower if you have more than one (unlikely).
	    }
#endif

    if (use_fast && (tty_usb==0))
      fd = rcx_init(tty, 1); //For now, run USB only slow mode.
    else
      fd = rcx_init(tty, 0);

    if (receive)
    {
	  if (argc) filename = argv[0];
      if (print) msg_receive(fd);
      else file_receive(fd, argv[0]);
    }
    else
      file_send(fd, argv[0]);

    rcx_close(fd);

    return 0;
}

